# SPDX-License-Identifier: Apache-2.0

import base64
import os
import uuid
import mimetypes
from datetime import datetime, timezone

import peewee
from peewee_migrate import Router
from playhouse.pool import PooledSqliteDatabase
from argon2 import PasswordHasher
from itsdangerous.url_safe import URLSafeSerializer

from pyles.settings import DATABASE, SECRET_KEY

db = PooledSqliteDatabase(
    DATABASE,
    max_connections=30,
    stale_timeout=300,
)
db.connect()

migrations = Router(db)

ph = PasswordHasher()


class BaseModel(peewee.Model):
    class Meta:
        database = db


class User(BaseModel):
    id = peewee.UUIDField(default=uuid.uuid4, primary_key=True)
    username = peewee.CharField(max_length=40, unique=True)
    is_admin = peewee.BooleanField(null=False, default=False)
    password = peewee.CharField()
    salt = peewee.CharField()

    @classmethod
    def new(cls, username: str, password: str, is_admin=False) -> "User":
        return cls.create(
            username=username,
            password=ph.hash(password),
            salt=os.urandom(16).hex(),
            is_admin=is_admin,
        )

    def set_password(self, password: str):
        """Sets a password, but doesn't save it.
        This also changes the user's token salt."""

        self.password = ph.hash(password)
        self.salt = os.urandom(16).hex()

        return self

    def verify_password(self, password: str):
        """Checks the user's password.
        Returns True if the password is valid, False otherwise.
        Does not raise an error."""

        try:
            return ph.verify(self.password, password)
        except:
            return False

    def get_token(self):
        if not self.salt:
            self.salt = os.urandom(16).hex()
            self.save()

        s = URLSafeSerializer(SECRET_KEY, salt=self.salt)
        return s.dumps(str(self.id))

    def verify_token(self, token: str):
        s = URLSafeSerializer(SECRET_KEY, salt=self.salt)

        try:
            value = s.loads(token)
            return value == str(self.id)
        except:
            return False


def random_url_id():
    return base64.b64encode(os.urandom(12), altchars=b"bB").decode("utf-8")


class File(BaseModel):
    id = peewee.AutoField()
    url_id = peewee.CharField(null=False, unique=True, default=random_url_id)
    filename = peewee.CharField(null=False)
    hash = peewee.CharField(null=False)
    content_type = peewee.CharField(null=False)
    created_at = peewee.DateTimeField(
        null=False, default=lambda: datetime.now(tz=timezone.utc)
    )
    expires_at = peewee.DateTimeField(null=True)

    user = peewee.ForeignKeyField(User)

    @property
    def path(self):
        ext = mimetypes.guess_extension(self.content_type, strict=False)
        if not ext:
            ext = ""

        return f"{self.url_id}{ext}"


class Tag(BaseModel):
    id = peewee.AutoField()
    name = peewee.CharField()

    user = peewee.ForeignKeyField(User)


class FileTag(BaseModel):
    file = peewee.ForeignKeyField(File)
    tag = peewee.ForeignKeyField(Tag)


with db:
    db.create_tables([User, File, Tag, FileTag])

migrations.migrator.create_model(User)
migrations.migrator.create_model(File)
migrations.migrator.create_model(Tag)
migrations.migrator.create_model(FileTag)

migrations.run()
